// This is the MEX wrapper for QUIC-mixed.C.  The algorithm is in QUIC-mixed.C.
// Sparse Inverse Covarnace Estimation with Multiple Graphs
//
//      min_{X,Y: X{r}+Y{r}>0 for all r} -\sum_{r=1}^k logdet( X{r}+Y{r} )+ 
//                                       \sum_{r=1}^k trace(X{r}+Y{r}, Sigma{r})+ 
//                                       \sum_{r=1}^k lambda*||X{r}||_1 + 
//                                       beta*||Y||_gr.
//
// Invocation form within Matlab:
// [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta)
// [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta, maxIter)
// [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta, maxIter, msg)
//
// Input Arguments: 
// 		Sigma_list -- A cell array, each Sigma_list{r} is the r-th sample covariance
// 		lambda -- The L1-regularization parameter
// 		beta -- The Group Lasso regularization parameter
// 		maxIter -- Maximum number of iterations
// 		msg -- 0: no output; 1: print obj for each iter; 2: development mode
//
// Output Arguments: 
// 		Xlist -- A cell array, each Xlist{r} is the sparse part of the graphical model
// 		Ylist -- A cell array, each Ylist{r} is the group-sparse part of the graphicaml model
// 		objlist -- objective function value at each iteration
// 		timelist -- cummulated run time 
//

#include <mex.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

extern "C"
void QUICmixed(int k, int p, double **Slist, double **Xlist, double **Ylist, double lambda, double beta, int maxIter, double *opt, double *cputime, int msg);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nrhs < 3) {
	mexErrMsgIdAndTxt("QUIC:arguments",
			  "Missing arguments, usage:\n"
		"      [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta)\n"
		"      [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta, maxIter)\n"
		"      [Xlist Ylist objlist timelist] = QUIC(Sigma_list, lambda, beta, maxIter, msg)\n"
		"      \n"
		"      Arguments: \n"
	    "             Sigma_list - the cell array of empirical covariance matrices\n"
		"             lambda - the regularization parameter for L1 norm \n"
		"             beta - the group lasso regularization\n"
		"             maxIter - maximum number of iterations\n"
		"             msg - 0: no output; 1: print obj for each iter; 2: development mode");
    }
    long argIdx = 0;
   
	if (!mxIsCell(prhs[argIdx]))
	{
		mexErrMsgIdAndTxt("QUIC:type", 
				"Expected a cell array. (Arg. %d)\n", 
				argIdx+1);
	}

	int numElem = mxGetNumberOfElements(prhs[0]);
	int p;

	double **buf = (double **)malloc(sizeof(double *)*numElem);
	for (int i=0 ; i<numElem ; i++ )
	{
		mxArray *tmp =  mxGetCell(prhs[0], i);
		if ( !mxIsDouble(tmp) )
		{
			mexErrMsgIdAndTxt("QUIC:type", 
					"Expected a double araay. (Arg. %d{%d})", 
					argIdx+1, i+1);
		}
		buf[i] = (double *)mxGetPr(tmp);
		if ( i==0 ) 
			p = mxGetM(tmp); 
		if ( p!= mxGetM(tmp) || p!=mxGetN(tmp) ) {
			mexErrMsgIdAndTxt("QUIC:type", 
					"Expected a %d by %d double array. (Arg. %d: Slist{%d})\n", 
					p, p, argIdx+1, i+1);
		}
	}

	printf("%d Graphical models, p = %d\n", numElem, p);
	
	argIdx++;

    // Read regularization parameter: lambda
    if (!mxIsDouble(prhs[argIdx])) {
		mexErrMsgIdAndTxt("QUIC:type",
			  "Expected a double scalar. (Arg. %d)\n",
			  argIdx + 1);
	}
    double lambda = mxGetScalar(prhs[argIdx]);
    argIdx++;

	// Read regularization parameter: beta
    if (!mxIsDouble(prhs[argIdx])) {
		mexErrMsgIdAndTxt("QUIC:type",
			  "Expected a double scalar. (Arg. %d)\n",
			  argIdx + 1);
	}
    double beta = mxGetScalar(prhs[argIdx]);
    argIdx++;

    // Maximum number of Newton ierations (whole matrix update):
    int maxIter = 10; // default 10
    if (nrhs > argIdx) {
		maxIter = (int) mxGetScalar(prhs[argIdx]);
		argIdx++;
    }

	// Mode, default 1
	int msg = 1; 
	if (nrhs > argIdx) {
		msg = (int) mxGetScalar(prhs[argIdx]);
		argIdx++;
	}

	// Output Xlist
	mwSize dims[] = {p,p};
	mwSize dimdims[] = {numElem};
	plhs[0] = mxCreateCellArray(1, dimdims);
	mxArray *cell_array_ptr = plhs[0];
	double **Xlist = (double **)malloc(sizeof(double *)*numElem);
	for ( int i=0 ; i<numElem ; i++ ) {
		mxArray *tmp = mxCreateDoubleMatrix(p, p, mxREAL);
		mxSetCell(cell_array_ptr, i, tmp);
		Xlist[i] = (double *)mxGetPr(tmp);
		for ( int j=0 ; j<p*p ; j++ )
			Xlist[i][j] = 0;
		// Use Identiy as initialization
		for ( int j = 0; j < p*p; j += (p+1)) {
		    Xlist[i][j] = 1.0;
		}
	}

	// Output Ylist
	plhs[1] = mxCreateCellArray(1, dimdims);
	mxArray *cell_array_ptr_1 = plhs[1];
	double **Ylist = (double **)malloc(sizeof(double *)*numElem);
	for ( int i=0 ; i<numElem ; i++ ) {
		mxArray *tmp = mxCreateDoubleMatrix(p, p, mxREAL);
		mxSetCell(cell_array_ptr_1, i, tmp);
		Ylist[i] = (double *)mxGetPr(tmp);
		for ( int j=0 ; j<p*p ; j++ )
			Ylist[i][j] = 0;
		for ( int j = 0; j < p*p; j += (p+1)) {
		    Ylist[i][j] = 1.0;
		}
	}

    double* opt = NULL;
    double* cputime = NULL;
    unsigned long optSize = maxIter;
    
	if (nlhs > 2) {
		plhs[2] = mxCreateDoubleMatrix(optSize, 1, mxREAL);
		opt = mxGetPr(plhs[2]);
    }
    if (nlhs > 3) {
		plhs[3] = mxCreateDoubleMatrix(optSize, 1, mxREAL);
		cputime = mxGetPr(plhs[3]);
    }

	QUICmixed(numElem, p, buf, Xlist, Ylist, lambda, beta, maxIter, opt, cputime, msg);
	free(buf);
}
