// Solves the regularized inverse covariance matrix selection using a
// combination of Newton's method, quadratic approximation and
// coordinate descent.  The code was written by Cho-Jui Hsieh. 

// Send questions, comments and license inquiries to: cjhsieh@cs.utexas.edu

#define VERSION "1.1"

#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#ifdef GDEBUG
#include "startgdb.h"
#endif

#ifdef LANG_R
  #include <R_ext/Print.h>
  #define MSG Rprintf
#endif
#if defined(LANG_M) || defined(MATLAB_MEX_FILE)
  #include <mex.h>
  #define MSG mexPrintf
#endif

#ifndef MSG
  #define MSG printf
#endif

#define EPS (double(2.22E-16))

#if !defined(LANG_R) && defined(_WIN32)
  #define dpotrf_ dpotrf
  #define dpotri_ dpotri
#endif

// It would be preferable to use an include such as lapack.h.  Except
// lapack.h is not available from the octave or liblapack-dev packages...
extern "C" {
    void dpotrf_(char* uplo, ptrdiff_t* n, double* A, ptrdiff_t* lda,
		 ptrdiff_t* info);
    void dpotri_(char* uplo, ptrdiff_t* n, double* A, ptrdiff_t* lda,
		 ptrdiff_t* info);
}

typedef struct {
    unsigned short i;
    unsigned short j;
} ushort_pair_t;

static inline void CoordinateDescentUpdate(
    unsigned long p, const double* const S, const double lambda,
    const double* X, const double* W, double* U, double* D,
    unsigned long i, unsigned long j, double& normD, double& diffD)
{
    unsigned long ip = i*p;
    unsigned long jp = j*p;
    unsigned long ij = ip + j;

    double a = W[ij]*W[ij];
    if (i != j)
        a += W[ip+i]*W[jp+j];
    double ainv = 1.0/a;  // multiplication is cheaper than division

    double b = S[ij] - W[ij];
    for (unsigned long k = 0; k < p ; k++)
        b += W[ip+k]*U[k*p+j];

    double l = lambda*ainv;
    double c = X[ij] + D[ij];
    double f = b*ainv;
    double mu;
    normD -= fabs(D[ij]);
    if (c > f) {
        mu = -f - l;
        if (c + mu < 0.0) {
            mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    } else {
	mu = -f + l;
	if (c + mu > 0.0) {
	    mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    }
    diffD += fabs(mu);
    normD += fabs(D[ij]);
    if (mu != 0.0) {
        for (unsigned long k = 0; k < p; k++)
            U[ip+k] += mu*W[jp+k];
        if (i != j) {
            for (unsigned long k = 0; k < p; k++)
                U[jp+k] += mu*W[ip+k];
        }
    }
}

static inline void CoordinateDescentUpdate_GroupLasso(
    int p, int k, double** Slist, const double beta,
    double** Ylist, double** Wlist, double** Ulist, double** DYlist,
    unsigned long i, unsigned long j,  double& normD, double& diffD)
{
    unsigned long ip = i*p;
    unsigned long jp = j*p;
    unsigned long ij = ip + j;

	double *grad = (double *)malloc(sizeof(double)*k);
	double *hessian = (double *)malloc(sizeof(double)*k);
	double *yorigin = (double *)malloc(sizeof(double)*k);
	double *dy = (double *)malloc(sizeof(double)*k);

	for ( int r=0 ; r<k ; r++ ) {
	    double a = Wlist[r][ij]*Wlist[r][ij];
	    if (i != j)
	        a += Wlist[r][ip+i]*Wlist[r][jp+j];
    	double b = Slist[r][ij] - Wlist[r][ij];
		for (unsigned long kk = 0; kk < p ; kk++)
			b += Wlist[r][ip+kk]*Ulist[r][kk*p+j];
		hessian[r] = a;
		grad[r] = b - a*(Ylist[r][ij]+ DYlist[r][ij]);
		yorigin[r] = Ylist[r][ij];
		dy[r] = DYlist[r][ij];
	}

	double gnorm = 0.0;
	for ( int r=0; r<k ; r++ ) 
		gnorm += grad[r]*grad[r];
	gnorm =sqrt(gnorm);

	if ( gnorm < beta ) {
		for ( int r=0 ; r<k ; r++ ) {
			double newdelta = -yorigin[r];
			if ( newdelta != dy[r]) {
				double dd = newdelta - dy[r];
				// Update Wlist[r]
		        for (unsigned long kk = 0; kk < p; kk++)
		            Ulist[r][ip+kk] += dd*Wlist[r][jp+kk];
		        if (i != j) {
		            for (unsigned long kk = 0; kk < p; kk++)
		                Ulist[r][jp+kk] += dd*Wlist[r][ip+kk];
		        }
				DYlist[r][ij] = newdelta;
				diffD += fabs(dd);
				normD += fabs(newdelta);
			}
		}
	} else {
		// Update
		double Delta = 0;
		for ( int r=0 ; r<k ; r++ )
			Delta += pow(yorigin[r]+dy[r],2);
		Delta = sqrt(Delta);
		int maxnewton = 10;
		for ( int newtoniter = 0 ; newtoniter < maxnewton ; newtoniter ++ )
		{
			double ynormsq = 0;
			for ( int r=0 ; r<k ; r++ )
				ynormsq += pow(grad[r]/(hessian[r]*Delta + beta),2);
			double dynorm = 0;
			for ( int r=0 ; r<k ; r++ )
				dynorm += (grad[r]*grad[r]*hessian[r])/pow(hessian[r]*Delta+beta ,3);
			dynorm*=(-2);
			double gradgrad = 0;
			gradgrad = -0.5*pow(ynormsq, -3.0/2.0)*dynorm;
			
			double fvalue = 1/sqrt(ynormsq) -1;
//			printf("newtoniter %d 1/||y||-1: %lf\n", newtoniter, fvalue);
			Delta = Delta - fvalue/gradgrad; 
			if ( fvalue < 1e-8)
				break;
		}
		for ( int r=0 ; r<k ; r++ )
		{
			double nowx =  -grad[r]/(hessian[r]+beta/Delta);
			double newdelta = nowx - yorigin[r];
			if ( newdelta != dy[r]) {
				double dd = newdelta - dy[r];
				// Update Wlist[r]
		        for (unsigned long kk = 0; kk < p; kk++)
		            Ulist[r][ip+kk] += dd*Wlist[r][jp+kk];
		        if (i != j) {
		            for (unsigned long kk = 0; kk < p; kk++)
		                Ulist[r][jp+kk] += dd*Wlist[r][ip+kk];
		        }
				DYlist[r][ij] = newdelta;
				diffD += fabs(dd);
				normD += fabs(newdelta);
			}
		}
	}

	free(grad);
	free(hessian);
	free(yorigin);
	free(dy);
}

double CompgroupnormX(double **Ylist, int p, int k)
{
	double groupnorm = 0.0;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=p )
		for ( int j=0 ; j<i ; j++ ) {
			double l2norm = 0.0;
			for ( int r=0 ; r<k ; r++ )
				l2norm += Ylist[r][kk+j]*Ylist[r][kk+j] ;
			groupnorm += sqrt(l2norm);
		}
	groupnorm *= 2;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=(p+1) ) {
		double l2norm = 0.0;
		for ( int r=0 ; r<k ; r++)
			l2norm += Ylist[r][kk]*Ylist[r][kk];
		groupnorm += sqrt(l2norm);
	}
	
	return groupnorm;
}



double Compl1normX(double *X, int p)
{
	double l1normX = 0.0;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=p )
		for ( int j=0 ; j<i ; j++ ) {
			l1normX += fabs(X[kk+j]);
		}
	l1normX *= 2.0;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=(p+1) ) {
		l1normX += fabs(X[kk]);
	}

	return l1normX;
}


double CompTrSX(double *S, double *X, int p)
{
	double trSX = 0.0;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=p )
		for ( int j=0 ; j<i ; j++ ) {
			trSX += S[kk+j]*X[kk+j];
		}
	trSX *= 2.0;
	for ( int i=0, kk=0 ; i<p ; i++, kk+=(p+1) ) {
		trSX += S[kk]*X[kk];
	}
	return trSX;
}

void CompApwB_upper(double *A, double *B, double *C, int p, double w) 
{
	for ( int i=0, kk=0 ; i<p ; i++, kk+=p) 
		for ( int j=0 ; j<=i ; j++)
			C[kk+j] = A[kk+j]+w*B[kk+j];
}


void CompApB_upper(double *A, double *B, double *C, int p) 
{
	for ( int i=0, kk=0 ; i<p ; i++, kk+=p) 
		for ( int j=0 ; j<=i ; j++)
			C[kk+j] = A[kk+j]+B[kk+j];
}

// Storing Cholesky factors in W
// Return -1 if any W is not positive definite, return 0 otherwise
int CompObj(double **Xlist, double **Ylist, double **Slist, double **Wlist, int p, int k, double lambda, double beta, double *value)
{
	double fX = 0.0;
	for ( int r=0 ; r<k ; r++ )
	{
		ptrdiff_t info = 0;
		ptrdiff_t p0 = p;
		double *U = Wlist[r];

		memset(U, 0, p*p*sizeof(double));
		CompApB_upper(Xlist[r], Ylist[r], U, p);
		double trSX = CompTrSX(Slist[r], U, p);

		dpotrf_((char*) "U", &p0, U, &p0, &info);

		if (info != 0) {
//			MSG("Lack of positive definiteness!\n");
			return -1;
		}

		double logdetX = 0.0;
		for (unsigned long i = 0, kk = 0; i < p; i++, kk += (p+1))
			logdetX += log(U[kk]);
		logdetX *= 2.0;

		double l1normX = lambda*Compl1normX(Xlist[r], p);
		fX = fX + (trSX + l1normX) - logdetX;
	}

	fX = fX + beta*CompgroupnormX(Ylist, p, k);
	*value = fX;

	return 1;
}

extern "C"
void QUICmixed(int k, int p, double **Slist, double **Xlist, double **Ylist, double lambda, double beta, int maxIter, double *opt, double *cputime, int msg)
{
    double timeBegin = clock();
    srand(1);
    unsigned long maxNewtonIter = maxIter;
    unsigned long max_lineiter = 20;
    double fX = 1e+15;
    double fX1 = 1e+15;
    double fXprev = 1e+15;
    double sigma = 0.001;
	double diffD;
	double logdetX = 0.0;
	double trSX = 0.0;

	// Construct arrays
	double **DXlist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		DXlist[r] = (double *)malloc(sizeof(double)*p*p);
	
	double **DYlist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		DYlist[r] = (double *)malloc(sizeof(double)*p*p);

	double **Ulist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		Ulist[r] = (double *)malloc(sizeof(double)*p*p);

	double **Wlist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		Wlist[r] = (double *)malloc(sizeof(double)*p*p);

	double **newXlist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		newXlist[r] = (double *)malloc(sizeof(double)*p*p);
	
	double **newYlist = (double **)malloc(sizeof(double *)*k);
	for ( int r=0 ; r<k ; r++ )
		newYlist[r] = (double *)malloc(sizeof(double)*p*p);


	int *numActive_X_list = (int *)malloc(sizeof(int)*k);
	int numActive_Y;

	ushort_pair_t* activeSet_Y = (ushort_pair_t*)malloc(p*(p+1)/2*sizeof(ushort_pair_t));
	ushort_pair_t** activeSet_X_list = (ushort_pair_t**)malloc(sizeof(ushort_pair_t*)*k);
	for ( int r=0 ; r<k ; r++ )
		activeSet_X_list[r] =  (ushort_pair_t*)malloc(p*(p+1)/2*sizeof(ushort_pair_t));

	// Initial Wlist -- assume initial X and Y are diagonal matrices
	// Wlist{r} = (Xlist{r}+Ylist{r})^(-1)
	for ( int r=0 ; r<k ; r++ ) {
		memset(Wlist[r], 0, p*p*sizeof(double));
		for ( int i = 0, kk = 0; i < p; i++, kk += (p+1))
			Wlist[r][kk] = 1.0/(Xlist[r][kk]+Ylist[r][kk]);
	}

    for ( int NewtonIter = 0; NewtonIter < maxIter; NewtonIter++ ) {

		// Initial DXlist, DYlist, Ulist
		for ( int r=0 ; r<k ; r++ ) {
			memset(DXlist[r], 0, p*p*sizeof(double));
			memset(DYlist[r], 0, p*p*sizeof(double));
			memset(Ulist[r], 0, p*p*sizeof(double));
		}

	    // Compute the active set on X:
		for ( int r=0 ; r<k ; r++ )
		    numActive_X_list[r] = 0;
		numActive_Y = 0;

		double subgrad = 0;
		for ( int r=0 ; r<k ; r++ ) {
			for ( int kk = 0, i = 0; i < p; i++, kk += p) {
				for ( int j = 0; j <= i; j++) {
					double g = Slist[r][kk+j] - Wlist[r][kk+j];
					if ( Xlist[r][k+j] != 0.0 || (fabs(g) > lambda) ) {
						activeSet_X_list[r][numActive_X_list[r]].i = (unsigned short) i;
						activeSet_X_list[r][numActive_X_list[r]].j = (unsigned short) j;
						numActive_X_list[r]++;
						if (Xlist[r][k+j] > 0) 
							g += lambda;
						else if (Xlist[r][k+j] < 0) 
							g -= lambda;
						else 
							g = fabs(g) - lambda;
						if ( i == j)
							subgrad += fabs(g);
						else
							subgrad += 2*fabs(g);
					}
				}
			}
		}

		for ( int kk=0, i=0 ; i<p ; i++, kk+=p ) {
			for ( int j=0 ; j<=i ; j++ ) {
				double g = 0;
				for ( int r=0 ; r<k ; r++ )
					g += (Slist[r][kk+j] - Wlist[r][kk+j])*(Slist[r][kk+j] - Wlist[r][kk+j]);
				g = sqrt(g);

				int iszero = 1;
				for ( int r=0 ; r<k ; r++ )
					if ( Xlist[r][kk+j] != 0.0 )
						iszero = 0;
				if ( iszero == 0 || (g > beta)) {
					activeSet_Y[numActive_Y].i = (unsigned short) i;
					activeSet_Y[numActive_Y].j = (unsigned short) j;
					numActive_Y++;
					if ( i==j)
						subgrad += fabs(g-beta);
					else
						subgrad += 2*fabs(g-beta);
				}
			}
		}

		int total_active_X = 0;
		for ( int r=0 ; r<k ; r++ )
			total_active_X += numActive_X_list[r];

		if (msg >= 2) {
			MSG("Newton iteration %d.\n", NewtonIter);
			MSG("  Active set size of X = %d / %d.\n", total_active_X, p*p*k);
			MSG("  Active set size of Y = %d / %d.\n", numActive_Y, p*p);
		}

		for (unsigned long cdSweep = 1; cdSweep <= 1+NewtonIter/3;
				cdSweep++) {
			
			diffD = 0.0;
			double normD = 0.0;
			
			// Update X
			for ( int r=0 ; r<k ; r++ ) {
				ushort_pair_t* activeSet = activeSet_X_list[r];
				for (int i = 0; i < numActive_X_list[r]; i++ ) {
					int j = i + rand()%(numActive_X_list[r] - i);
					unsigned short k1 = activeSet[i].i;
					unsigned short k2 = activeSet[i].j;
					activeSet[i].i = activeSet[j].i;
					activeSet[i].j = activeSet[j].j;
					activeSet[j].i = k1;
					activeSet[j].j = k2;
				}
				
				for (int l = 0; l < numActive_X_list[r]; l++) {
					unsigned long i = activeSet[l].i;
					unsigned long j = activeSet[l].j;
					CoordinateDescentUpdate(p, Slist[r], lambda, Xlist[r], Wlist[r],
							Ulist[r], DXlist[r], i, j, normD, diffD);
				}
			}

			// Update Y

			for (int i = 0; i < numActive_Y; i++ ) {
				int j = i + rand()%(numActive_Y - i);
				unsigned short k1 = activeSet_Y[i].i;
				unsigned short k2 = activeSet_Y[i].j;
				activeSet_Y[i].i = activeSet_Y[j].i;
				activeSet_Y[i].j = activeSet_Y[j].j;
				activeSet_Y[j].i = k1;
				activeSet_Y[j].j = k2;
			}
				
			for (unsigned long l = 0; l < numActive_Y; l++) {
				unsigned long i = activeSet_Y[l].i;
				unsigned long j = activeSet_Y[l].j;
				CoordinateDescentUpdate_GroupLasso(p, k, Slist, beta, Ylist, Wlist,
							Ulist, DYlist, i, j, normD, diffD);
			}

			if (msg >= 2) {
				MSG("  Coordinate descent sweep %ld. norm of D = %e, "
						"change in D = %e.\n", cdSweep, normD, diffD);
			}
		}
		
		if (fX == 1e+15) {
			int info = CompObj(Xlist, Ylist, Slist, Wlist, p, k, lambda, beta, &fX);
			if ( info == -1 ) {
				if ( msg >=2 )
					printf("Lack of positive definite. \n");
				return;
			}
		}

		double alpha = 1.0;
		double l1normXD = 0.0;
		double fX1prev = 1e+15;
		for (unsigned long lineiter = 0; lineiter < max_lineiter;
				lineiter++) {

			for ( int r=0 ; r<k ; r++ ) {
				CompApwB_upper(Xlist[r], DXlist[r], newXlist[r], p, alpha);
				CompApwB_upper(Ylist[r], DYlist[r], newYlist[r], p, alpha);
			}

			// Wlist hold the Cholesky factors
			int info = CompObj(newXlist, newYlist, Slist, Wlist, p, k, lambda, beta, &fX1);
			if ( info ==-1 ) {
				alpha *= 0.5;
				continue;
			}

			if (fX1 < fX) {
				if (msg >= 2)
					MSG("    Line search step size chosen: %e.\n", alpha);
				fXprev = fX;
				fX = fX1;
				break;
			}
			
			if (msg >= 2) {
				MSG("    Line search step size %e.\n", alpha);
				MSG("      Objective value would not decrease sufficiently: "
						"%e.\n", fX1 - fX);
			}
			alpha *= 0.5;
		}
		
		// compute W = inv(X):
		for ( int r=0 ; r<k ; r++ ) {
			ptrdiff_t info;
			ptrdiff_t p0 = p;
			dpotri_((char*) "U", &p0, Wlist[r], &p0, &info);

			for (unsigned long i = 0; i < p; i++) {
				for (unsigned long j = 0; j <= i; j++) {
					double tmp = Wlist[r][i*p+j];
					Wlist[r][j*p+i] = tmp;
				}
			}
			// SWAP Xlist and newXlist; Ylist and newYlist

			double **tmp = Xlist;
			Xlist = newXlist;
			newXlist = tmp;
			tmp = Ylist;
			Ylist = newYlist;
			newYlist = tmp;
		}

		if (opt != NULL)
			opt[NewtonIter] = fX;
		double nowtime = (clock()-timeBegin)/CLOCKS_PER_SEC;
		if (time != NULL) {
		// Note that nn trace mode we do not want to count the
		// time spent computing the duality gap.
			cputime[NewtonIter] = nowtime;
	    }

		if ( msg>=1 ) {
			MSG("Iter %d time %lf obj %lf.\n", NewtonIter, nowtime, fX);
		}

    }

	for ( int r=0 ; r<k ; r++ )
		free(DXlist[r]);
	free(DXlist);
	for ( int r=0 ; r<k ; r++ )
		free(DYlist[r]);
	free(DYlist);
	for ( int r=0 ; r<k ; r++ )
		free(Ulist[r]);
	free(Ulist);
	for ( int r=0 ; r<k ; r++ )
		free(Wlist[r]);
	for ( int r=0 ; r<k ; r++ )
		free(newXlist[r]);
	free(newXlist);
	for ( int r=0 ; r<k ; r++ )
		free(newYlist[r]);
	free(newYlist);
	free(Wlist);
	free(numActive_X_list);
	free(activeSet_Y);
	for ( int r=0 ; r<k ; r++ )
		free(activeSet_X_list[r]);
	free(activeSet_X_list);

	// Fill the lower triangular part of  X
	for ( int r=0 ; r<k ; r++ )
	    for (unsigned long i = 0, k = 0; i < p; i++, k += p)
			for (unsigned long j = i+1; j < p; j++)
				Xlist[r][k+j] = Xlist[r][j*p+i];

	for ( int r=0 ; r<k ; r++ )
	    for (unsigned long i = 0, k = 0; i < p; i++, k += p)
			for (unsigned long j = i+1; j < p; j++)
				Ylist[r][k+j] = Ylist[r][j*p+i];

    double elapsedTime = (clock() - timeBegin)/CLOCKS_PER_SEC;
    
//	if (msg >= QUIC_MSG_MIN)
	if (1)
	MSG("QUIC CPU time: %.3f seconds\n", elapsedTime);
}
