// Solves the regularized inverse covariance matrix selection using a
// combination of Newton's method, quadratic approximation and
// coordinate descent.  The original algorithm was coded by Cho-Jui
// Hsieh.  Improvements were made by Matyas A. Sustik.
// This code is released under the GPL version 3.
// 
// Added the new function  
// void QUIC_quad(int& p, const double* Sigma, double& lambda,
//          	  int& msg, double* X, double* W, int maxit)
// To conduct coordinate descent update with a fixed Hessian

#define VERSION "1.1"

#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>
#ifdef GDEBUG
#include "startgdb.h"
#endif

#ifdef LANG_R
  #include <R_ext/Print.h>
  #define MSG Rprintf
#endif
#if defined(LANG_M) || defined(MATLAB_MEX_FILE)
  #include <mex.h>
  #define MSG mexPrintf
#endif

#ifndef MSG
  #define MSG printf
#endif

#define EPS (double(2.22E-16))
//#define EPS ((double)0x3cb0000000000000)

#if !defined(LANG_R) && defined(_WIN32)
  #define dpotrf_ dpotrf
  #define dpotri_ dpotri
#endif

// It would be preferable to use an include such as lapack.h.  Except
// lapack.h is not available from the octave or liblapack-dev packages...
extern "C" {
    void dpotrf_(char* uplo, ptrdiff_t* n, double* A, ptrdiff_t* lda,
		 ptrdiff_t* info);
    void dpotri_(char* uplo, ptrdiff_t* n, double* A, ptrdiff_t* lda,
		 ptrdiff_t* info);
}

typedef struct {
    unsigned short i;
    unsigned short j;
} ushort_pair_t;

static inline unsigned long IsDiag(unsigned long p, const double* A)
{
    for (unsigned long k = 0, i = 0; i < p; i++, k += p)
	for (unsigned long j = 0; j < i; j++)
	    if (A[k + j] != 0.0)
		return 0;
    return 1;
}

static inline void CoordinateDescentUpdate_test(
    unsigned long p, const double* const S, const double lambda,
    const double* X, const double* W, double* U, double* D,
    unsigned long i, unsigned long j, double& normD, double& diffD, const double* L)
{
    unsigned long ip = i*p;
    unsigned long jp = j*p;
    unsigned long ij = ip + j;

    double a = W[ij]*W[ij];
    if (i != j)
        a += W[ip+i]*W[jp+j];
    double ainv = 1.0/a;  // multiplication is cheaper than division

    double b = S[ij] - L[ij];
    for (unsigned long k = 0; k < p ; k++)
        b += W[ip+k]*U[k*p+j];

    double l = lambda*ainv;
    double c = X[ij] + D[ij];
    double f = b*ainv;
    double mu;

    normD -= fabs(D[ij]);
    if (c > f) {
        mu = -f - l;
        if (c + mu < 0.0) {
            mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    } else {
	mu = -f + l;
	if (c + mu > 0.0) {
	    mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    }
    diffD += fabs(mu);
    normD += fabs(D[ij]);
    if (mu != 0.0) {
        for (unsigned long k = 0; k < p; k++)
            U[ip+k] += mu*W[jp+k];
        if (i != j) {
            for (unsigned long k = 0; k < p; k++)
                U[jp+k] += mu*W[ip+k];
        }
    }
}

static inline void CoordinateDescentUpdate(
    unsigned long p, const double* const S, const double lambda,
    const double* X, const double* W, double* U, double* D,
    unsigned long i, unsigned long j, double& normD, double& diffD)
{
    unsigned long ip = i*p;
    unsigned long jp = j*p;
    unsigned long ij = ip + j;

    double a = W[ij]*W[ij];
    if (i != j)
        a += W[ip+i]*W[jp+j];
    double ainv = 1.0/a;  // multiplication is cheaper than division

    double b = S[ij] - W[ij];
    for (unsigned long k = 0; k < p ; k++)
        b += W[ip+k]*U[k*p+j];

    double l = lambda*ainv;
    double c = X[ij] + D[ij];
    double f = b*ainv;
    double mu;
    normD -= fabs(D[ij]);
    if (c > f) {
        mu = -f - l;
        if (c + mu < 0.0) {
            mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    } else {
	mu = -f + l;
	if (c + mu > 0.0) {
	    mu = -c;
	    D[ij] = -X[ij];
	} else {
	    D[ij] += mu;
	}
    }
    diffD += fabs(mu);
    normD += fabs(D[ij]);
    if (mu != 0.0) {
        for (unsigned long k = 0; k < p; k++)
            U[ip+k] += mu*W[jp+k];
        if (i != j) {
            for (unsigned long k = 0; k < p; k++)
                U[jp+k] += mu*W[ip+k];
        }
    }
}

// Return the objective value.
static inline double DiagNewton(unsigned long p, const double* S,
				const double* Lambda, const double* X,
				const double* W, double* D)
{
    for (unsigned long ip = 0, i = 0; i < p; i++, ip += p) {
	for (unsigned long jp = 0, j = 0; j < i; j++, jp += p) {
	    unsigned long ij = ip + j;
	    double a = W[ip + i]*W[jp + j];
	    double ainv = 1.0/a;  // multiplication is cheaper than division
	    double b = S[ij];
	    double l = Lambda[ij]*ainv;
	    double f = b*ainv;
	    double mu;
	    double x = -b*ainv;
	    if (0 > f) {
		mu = -f - l;
		x -= l;
		if (mu < 0.0) {
		    mu = 0.0;
		    D[ij] = -X[ij];
		} else {
		    D[ij] += mu;
		}
	    } else {
		mu = -f + l;
		if (mu > 0.0) {
		    mu = 0.0;
		    D[ij] = -X[ij];
		} else {
		    D[ij] += mu;
		}
	    }
	}
    }
    double logdet = 0.0;
    double l1normX = 0.0;
    double trSX = 0.0;
    for (unsigned long i = 0, k = 0; i < p; i++, k += (p+1)) {
	logdet += log(X[k]);
	l1normX += fabs(X[k])*Lambda[k];
	trSX += X[k]*S[k];
	double a = W[k]*W[k];
	double ainv = 1.0/a;  // multiplication is cheaper than division
	double b = S[k] - W[k];
	double l = Lambda[k]*ainv;
	double c = X[k];
	double f = b*ainv;
	double mu;
	if (c > f) {
	    mu = -f - l;
	    if (c + mu < 0.0) {
		D[k] = -X[k];
		continue;
	    }
	} else {
	    mu = -f + l;
	    if (c + mu > 0.0) {
		D[k] = -X[k];
		continue;
	    }
	}
	D[k] += mu;
    }
    double fX = -logdet + trSX + l1normX;
    return fX;
}

static double projLogDet(unsigned long p, const double* S,
			 double* W, double* prW, double* Lambda)
{
    // The computed W does not satisfy |W - S| .< Lambda.  Project it.
    for (unsigned long i = 0; i < p; i++) {
	for (unsigned long j = 0; j <= i; j++) {
	    double tmp = W[i*p+j];
	    if (S[i*p+j] - Lambda[i*p+j] > tmp)
		tmp = S[i*p+j] - Lambda[i*p+j];
	    if (S[i*p+j] + Lambda[i*p+j] < tmp)
		tmp = S[i*p+j] + Lambda[i*p+j];
	    prW[i*p+j] = tmp;
	    prW[j*p+i] = tmp;
	}
    }
    ptrdiff_t info = 0;
    ptrdiff_t p0 = p;
    dpotrf_((char*) "U", &p0, prW, &p0, &info);
    if (info != 0)
	return 1e+15;
    double logdet = 0.0;
    for (unsigned long i = 0, k = 0; i < p; i++, k += (p+1))
	logdet += log(prW[k]);
    logdet += logdet;
    return logdet;
}

#define QUIC_MSG_NO      0
#define QUIC_MSG_MIN     1
#define QUIC_MSG_NEWTON  2
#define QUIC_MSG_CD      3
#define QUIC_MSG_LINE    4

extern "C"
void QUIC_quad(int& p, const double* Sigma, double& lambda, 
	  int& msg, double* X, double* W, int maxit)
{
    double timeBegin = clock();
    srand(1);
    unsigned long maxNewtonIter = 1;
    double cdSweepTol = 0.05;
    unsigned long max_lineiter = 20;
    double fX = 1e+15;
    double fX1 = 1e+15;
    double fXprev = 1e+15;
    double sigma = 0.001;
    double* D = (double*) malloc(p*p*sizeof(double));
    double* U = (double*) malloc(p*p*sizeof(double));

	ushort_pair_t* activeSet = (ushort_pair_t*)malloc(p*(p+1)/2*sizeof(ushort_pair_t));
    double l1normX = 0.0;
    double trSX = 0.0;
    double logdetX = 0.0;
    for (unsigned long i = 0, k = 0; i < p ; i++, k += p) {
		for (unsigned long j = 0; j < i; j++) {
			l1normX += lambda*fabs(X[k+j]);
			trSX += X[k+j]*Sigma[k+j];
		}
    }
    l1normX *= 2.0;
    trSX *= 2.0;
    for (unsigned long i = 0, k = 0; i < p; i++, k += (p+1)) {
		l1normX += lambda*fabs(X[k]);
		trSX += X[k]*Sigma[k];
    }
    unsigned long pathIdx = 0;
    unsigned long NewtonIter = 1;
    for (; NewtonIter <= maxNewtonIter; NewtonIter++) {
		double normD = 0.0;
		double diffD = 0.0;
		double subgrad = 1e+15;	

	    unsigned long numActive = 0;
	    memset(U, 0, p*p*sizeof(double));
		memset(D, 0, p*p*sizeof(double));
		subgrad = 0.0;
		for (unsigned long k = 0, i = 0; i < p; i++, k += p) {
			for (unsigned long j = 0; j <= i; j++) {
				double g = Sigma[k+j] - W[k+j];
				if (X[k+j] != 0.0 || (fabs(g) > lambda)) {
					activeSet[numActive].i = (unsigned short) i;
					activeSet[numActive].j = (unsigned short) j;
					numActive++;
					if (X[k+j] > 0) 
						g += lambda;
					else if (X[k+j] < 0) 
						g -= lambda;
					else 
						g = fabs(g) - lambda;
					subgrad += fabs(g);
				}
			}
		}
		if (msg >= QUIC_MSG_NEWTON) {
			MSG("Newton iteration %ld.\n", NewtonIter);
			MSG("  Active set size = %ld.\n", numActive);
			MSG("  sub-gradient = %e, l1-norm of X = %e.\n",
					subgrad, l1normX);
		}


		for (unsigned long cdSweep = 1; cdSweep <= maxit;
				cdSweep++) {
			diffD = 0.0;

			for (unsigned long i = 0; i < numActive; i++ ) {
				unsigned long j = i + rand()%(numActive - i);
				unsigned short k1 = activeSet[i].i;
				unsigned short k2 = activeSet[i].j;
				activeSet[i].i = activeSet[j].i;
				activeSet[i].j = activeSet[j].j;
				activeSet[j].i = k1;
				activeSet[j].j = k2;
			}
			for (unsigned long l = 0; l < numActive; l++) {
				unsigned long i = activeSet[l].i;
				unsigned long j = activeSet[l].j;
				CoordinateDescentUpdate(p, Sigma, lambda, X, W,
						U, D, i, j, normD, diffD);
			}
			if (msg >= QUIC_MSG_CD) {
				MSG("  Coordinate descent sweep %ld. norm of D = %e, "
						"change in D = %e.\n", cdSweep, normD, diffD);
			}
			if (diffD <= normD*cdSweepTol)
				break;
		}
		
		for (unsigned long i = 0, k = 0; i < p; i++, k += p)
			for (unsigned long j = 0; j <= i; j++)
				X[k+j] += D[k+j];
		break;
    }

	free(activeSet);
    free(U);
    free(D);
    for (unsigned long i = 0, k = 0; i < p; i++, k += p)
		for (unsigned long j = i+1; j < p; j++)
			X[k+j] = X[j*p+i];
    double elapsedTime = (clock() - timeBegin)/CLOCKS_PER_SEC;
    
	if (msg >= QUIC_MSG_MIN)
	MSG("QUIC CPU time: %.3f seconds\n", elapsedTime);
}


