// This is the MEX wrapper for QUICquad.  The algorithm is in QUIC.C.
// Solving min_{X:X-L>0} -logdet(X-L)+trace(X, Sigma)+alpha*||X||_1
// Invocation form within Matlab or Octave:
// [X] = QUICquad(Sigma, alpha, msg, maxIter, X0, W0)
//

#include <mex.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <stdint.h>

extern "C"
void QUIC_quad(int& p, const double* Sigma, double& lambda, 
	  int& msg, double* X, double* W, int maxit);

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[])
{
    if (nrhs < 3) {
	mexErrMsgIdAndTxt("QUIC:arguments",
			  "Missing arguments, please specify\n"
	    "             Sigma - the empirical covariance matrix, and\n"
	    "             L - the latent factor." 
		"             alpha - the regularization parameter");
    }
    long argIdx = 0;
    
	// The empirical covariance matrix:
    if (!mxIsDouble(prhs[argIdx])) 
	{
		mexErrMsgIdAndTxt("QUIC:type",
			  "Expected a double matrix. (Arg. %d)",
			  argIdx + 1);
	}
    const double* Sigma = mxGetPr(prhs[argIdx]);
    int p = (int) mxGetN(prhs[argIdx]);
    if (p != int(mxGetM(prhs[argIdx]))) {
	mexErrMsgIdAndTxt("QUIC:dimensions",
			  "Expected a square empirical covariance matrix.");
    }
    argIdx++;


    // Regularization parameter:
    if (!mxIsDouble(prhs[argIdx]))
	mexErrMsgIdAndTxt("QUIC:type",
			  "Expected a double scalar. (Arg. %d)",
			  argIdx + 1);
    double alpha = mxGetScalar(prhs[argIdx]);
    argIdx++;

    int msg = 0;    
    if (nrhs > argIdx) {
	msg = (int) mxGetScalar(prhs[argIdx]);
	argIdx++;
    }
    
    // Maximum number of Newton ierations (whole matrix update):
    int maxIter = 50;
    if (nrhs > argIdx) {
	maxIter = (int) mxGetScalar(prhs[argIdx]);
	argIdx++;
    }

    double* X0 = NULL;
    double* W0 = NULL;
    if (nrhs > argIdx) {
	if (!mxIsDouble(prhs[argIdx]))
	    mexErrMsgIdAndTxt("QUIC:type",
			      "Expected a double matrix. (Arg. %d)",
			      argIdx + 1);
	if (p != mxGetM(prhs[argIdx]) || p != mxGetN(prhs[argIdx]))
	    mexErrMsgIdAndTxt("QUIC:dimensions",
			      "Matrix dimensions should match.\n"
	         "             Maybe incorrect mode is specified?");
	X0 = mxGetPr(prhs[argIdx]);
	argIdx++;
	if (nrhs == argIdx)
	    mexErrMsgIdAndTxt("QUIC:initializations",
			      "Please specify both the initial estimate\n"
	         "             and the inverse.\n"
	         "             Maybe incorrect mode is specified?");
	if (!mxIsDouble(prhs[argIdx]))
	    mexErrMsgIdAndTxt("QUIC:type",
			      "Expected a double matrix. (Arg. %d)",
			      argIdx + 1);
	if (p != mxGetM(prhs[argIdx]) || p != mxGetN(prhs[argIdx])) {
	    mexErrMsgIdAndTxt("QUIC:dimensions",
			      "Matrix dimensions should match.\n"
	         "             Maybe incorrect mode is specified?");
	}
	W0 = mxGetPr(prhs[argIdx]);
	argIdx++;
    }

    double* X = NULL;
	mwSize dims[] = {p, p};
	mxArray* tmp = mxCreateNumericArray(2, dims, mxDOUBLE_CLASS, mxREAL);
	X = (double*) mxGetPr(tmp);
	if (nlhs > 0)
	    plhs[0] = tmp;
    
	if (X0 != NULL) {
	memcpy(X, X0, sizeof(double)*p*p);
    } else {
	memset(X, 0, sizeof(double)*p*p);
	for (unsigned long i = 0; i < p*p; i += (p+1)) {
	    X[i] = 1.0;
	}
	printf("WARNING: didn't specify initial. \n");
    }


    double* opt = NULL;
    double* cputime = NULL;
    int* iter = NULL;
    unsigned long optSize = maxIter;
    unsigned long iterSize = 1;
    

	QUIC_quad(p, Sigma,  alpha, msg,  X, W0, maxIter);
}
